@@ -7,7 +7,7 @@ use Mojo::JSON;
 use Mojo::Util;
 use constant DEBUG => $ENV{MOJO_OPENAPI_DEBUG} || 0;
 
-our $VERSION = '2.12';
+our $VERSION = '2.13';
 my $X_RE = qr{^x-};
 
 has route     => sub {undef};
@@ -349,14 +349,14 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
   }, "echo";
 
   # Load specification and start web server
-  plugin OpenAPI => {url => "data:///api.json"};
+  plugin OpenAPI => {url => "data:///spec.json"};
   app->start;
 
   __DATA__
-  @@ api.json
+  @@ spec.json
   {
     "swagger" : "2.0",
-    "info" : { "version": "0.8", "title" : "Pets" },
+    "info" : { "version": "0.8", "title" : "Echo Service" },
     "schemes" : [ "http" ],
     "basePath" : "/api",
     "paths" : {
